#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, Final, Optional, Protocol

from pdm_pfsc.config import (
    ConfigAccessor,
    ConfigItems,
    ConfigSection,
    IsMissing,
)
from pdm_pfsc.logging import traced_function

if TYPE_CHECKING:
    import argparse
    from collections.abc import Iterable, Mapping
    from pathlib import Path
    from types import SimpleNamespace

    from pdm.project.config import ConfigItem
    from pdm_pfsc.abstractions import ConfigHolder
    from pdm_pfsc.config import Config as SupportsConfigFile


class _SupportsAddConfigItems(Protocol):  # pylint: disable=R0903
    """"""

    def add_config(self, name: str, value: "ConfigItem") -> None:
        """"""
        raise NotImplementedError()


class _StringEnum(str, Enum):
    """"""

    # Justification: Zen of python: Explicit is better than implicit
    pass  # pylint: disable=W0107


class _ConfigKeys(_StringEnum):
    """"""

    VERSION = "version"
    VERSION_SOURCE = "source"
    VERSION_SOURCE_FILE_PATH = "path"
    BUILD_BACKEND = "build-backend"
    VCS_PROVIDER = "provider"
    PROJECT_METADATA = "project"


VERSION_CONFIG_KEY_NAME: "Final[str]" = _ConfigKeys.VERSION

COMMIT_MESSAGE_TEMPLATE_DEFAULT: "Final[str]" = (
    "chore: Bump version {from} to {to}\n\n"
    "Created a commit with a new version {to}.\n"
    "Previous version was {from}."
)
PERFORM_COMMIT_DEFAULT: "Final[bool]" = False
AUTO_TAG_DEFAULT: "Final[bool]" = False
TAG_ADD_PREFIX_DEFAULT: "Final[bool]" = True
ALLOW_DIRTY_DEFAULT: "Final[bool]" = False
VCS_PROVIDER_DEFAULT: "Final[str]" = "git-cli"


class _ConfigValues(_StringEnum):
    """"""

    VERSION_SOURCE_FILE = "file"
    VERSION_SOURCE_SCM = "scm"
    DEPRECATED_BUILD_BACKEND_PDM_PEP517_API = "pdm.pep517.api"
    BUILD_BACKEND_PDM_BACKEND = "pdm.backend"


class PdmBumpConfig:
    """"""

    def __init__(self, accessor: "_PdmBumpConfigAccessor") -> None:
        namespace = accessor.values
        self.__vcs_provider: str = namespace.vcs_provider
        self.__commit_msg_tpl: str = namespace.commit_msg_tpl
        self.__perform_commit: bool = namespace.perform_commit
        self.__auto_tag: bool = namespace.auto_tag
        self.__tag_add_prefix: bool = namespace.tag_add_prefix
        self.__allow_dirty: bool = namespace.allow_dirty

    @property
    def vcs_provider(self) -> "str":
        """"""
        return self.__vcs_provider

    @property
    def commit_msg_tpl(self) -> str:
        """"""
        return self.__commit_msg_tpl

    @property
    def perform_commit(self) -> bool:
        """"""
        return self.__perform_commit

    @property
    def auto_tag(self) -> bool:
        """"""
        return self.__auto_tag

    @property
    def tag_allow_dirty(self) -> bool:
        """"""
        return self.__allow_dirty

    @property
    def tag_add_prefix(self) -> bool:
        """"""
        return self.__tag_add_prefix

    def add_values_missing_in_cli(
        self, args: "argparse.Namespace"
    ) -> "argparse.Namespace":
        """"""

        args = PdmBumpConfig.__update_args(args, "commit", self.perform_commit)
        args = PdmBumpConfig.__update_args(
            args, "commit_message", self.commit_msg_tpl
        )
        args = PdmBumpConfig.__update_args(args, "tag", self.auto_tag)
        args = PdmBumpConfig.__update_args(args, "dirty", self.tag_allow_dirty)
        args = PdmBumpConfig.__update_args(
            args, "prepend_letter_v", self.tag_add_prefix
        )

        return args

    @staticmethod
    def __update_args(
        ns: "argparse.Namespace", key: str, value: Any
    ) -> "argparse.Namespace":
        if key not in ns:
            setattr(ns, key, value)
        else:
            stored_value: Optional[Any] = getattr(ns, key)
            if stored_value is None and value is not None:
                setattr(ns, key, value)
            elif isinstance(stored_value, IsMissing):
                if value is None:
                    setattr(ns, key, stored_value.raw_value())
                else:
                    setattr(ns, key, value)

        return ns


class _PdmBackendConfig:
    """"""

    def __init__(self, accessor: "ConfigAccessor") -> None:
        """"""
        self.__mapping = accessor.get_pyproject_config(
            ConfigSection.TOOL_CONFIG
        )

    @property
    def use_scm(self) -> bool:
        """"""
        return (
            self.__mapping.get_config_value(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_SCM
        )

    @property
    def use_file(self) -> bool:
        """"""
        return (
            self.__mapping.get_config_value(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_FILE
        )


class _BuildSystemConfig:
    """"""

    def __init__(self, accessor: "ConfigAccessor") -> None:
        """"""
        self.__accessor = accessor
        self.__mapping = accessor.get_pyproject_config(
            ConfigSection.BUILD_SYSTEM
        )
        self.__tool_mapping = accessor.get_pyproject_config(
            ConfigSection.TOOL_CONFIG
        )
        self.__backend = _PdmBackendConfig(accessor)

    @property
    def build_backend(self) -> str:
        """"""
        return self.__mapping.get_config_value(_ConfigKeys.BUILD_BACKEND)

    @property
    def uses_deprecated_build_backed_pdm_pep517(self) -> bool:
        """"""
        return (
            self.build_backend
            == _ConfigValues.DEPRECATED_BUILD_BACKEND_PDM_PEP517_API
            and self.__accessor.get_pyproject_tool_config(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_FILE
        )

    @property
    def uses_pdm_backend(self) -> bool:
        """"""
        return self.build_backend == _ConfigValues.BUILD_BACKEND_PDM_BACKEND

    @property
    def pdm_backend(self) -> "_PdmBackendConfig":
        """"""
        return self.__backend

    @property
    def version_source_file(self) -> "Optional[str]":
        """"""
        return self.__tool_mapping.get_config_value(
            _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE_FILE_PATH
        )


class _MetaDataConfig:
    """"""

    def __init__(self, accessor: "ConfigAccessor") -> None:
        """"""
        self.__accessor = accessor
        self.__build_system = _BuildSystemConfig(accessor)

    @property
    def version(self) -> "Optional[str]":
        """"""
        return self.__accessor.get_pyproject_metadata(_ConfigKeys.VERSION)

    @property
    def build_system(self) -> "_BuildSystemConfig":
        """"""
        return self.__build_system

    @property
    def is_dynamic_version(self) -> bool:
        """"""
        return _ConfigKeys.VERSION in self.__accessor.meta_data.dynamic


class _PdmBumpConfigAccessor(ConfigAccessor):
    def __init__(
        self,
        config: "SupportsConfigFile",
        cfg_holder: "ConfigHolder",
    ) -> None:
        super().__init__(config, cfg_holder)
        self.__items = ConfigItems(self)
        self.__items.add_config_value(
            "commit_msg_tpl",
            description=(
                "The default commit message. "
                "Uses templates 'from' and 'to'."
            ),
            default=COMMIT_MESSAGE_TEMPLATE_DEFAULT,
        )
        self.__items.add_config_value(
            "perform_commit",
            description=(
                "If set to true, commit the " "bumped changes automatically"
            ),
            default=PERFORM_COMMIT_DEFAULT,
        )
        self.__items.add_config_value(
            "auto_tag",
            description=(
                "Create a tag after bumping " "and committing the changes"
            ),
            default=AUTO_TAG_DEFAULT,
        )
        self.__items.add_config_value(
            "tag_add_prefix",
            description="Adds the prefix v to the version tag",
            default=TAG_ADD_PREFIX_DEFAULT,
        )
        self.__items.add_config_value(
            "allow_dirty",
            description="Allows tagging the project, if it is dirty",
            default=ALLOW_DIRTY_DEFAULT,
        )
        self.__items.add_config_value(
            "vcs_provider",
            description="Configures the VCS Provider to use.",
            default=VCS_PROVIDER_DEFAULT,
            use_env_var=True,
        )

    @property
    def plugin_config_name(self) -> "Iterable[str]":
        return {"bump"}

    @property
    def values(self) -> "SimpleNamespace":
        """"""
        return self.__items.to_namespace()

    def propagate(self, target: "_SupportsAddConfigItems") -> None:
        """"""
        self.__items.propagate(target)


class _EmptyConfig:
    @property
    def root(self) -> "Path":
        """"""
        raise NotImplementedError()

    @root.setter
    def root(self, value: "Path") -> None:
        """"""
        raise NotImplementedError()

    # Justification: Fulfill protocol
    @property
    # pylint: disable=C0103
    def PYPROJECT_FILENAME(self) -> str:  # noqa: N802 NOSONAR
        """"""
        raise NotImplementedError()

    # Justification: Fulfill protocol
    @PYPROJECT_FILENAME.setter
    # pylint: disable=C0103
    def PYPROJECT_FILENAME(self, value: str) -> None:  # noqa: N802 NOSONAR
        """"""
        raise NotImplementedError()

    @property
    def config(self) -> "Mapping[str, Any]":
        """"""
        return {}


# Justification: Just a protocol implementation
class _DummyConfig:  # pylint: disable=R0903
    @property
    def pyproject_file(self) -> "Path":
        """"""
        raise NotImplementedError()


class Config:
    """"""

    def __init__(self, project: "ConfigHolder") -> None:
        self.__project: "ConfigHolder" = project
        accessor: "_PdmBumpConfigAccessor" = _PdmBumpConfigAccessor(
            self, project
        )
        self.__pdm_bump = PdmBumpConfig(accessor)
        self.__meta_data = _MetaDataConfig(accessor)

    @staticmethod
    def get_config() -> "_PdmBumpConfigAccessor":
        """"""
        config: "ConfigHolder" = _EmptyConfig()
        accessor: "_PdmBumpConfigAccessor" = _PdmBumpConfigAccessor(
            _DummyConfig(), config
        )

        return accessor

    @property
    def pdm_bump(self) -> "PdmBumpConfig":
        """"""
        return self.__pdm_bump

    @property
    def meta_data(self) -> "_MetaDataConfig":
        """"""
        return self.__meta_data

    @cached_property
    @traced_function
    def pyproject_file(self) -> "Path":
        """"""
        return self.__project.root / self.__project.PYPROJECT_FILENAME



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

""""""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from difflib import unified_diff
from functools import cached_property
from pathlib import Path
from re import M as MultilinePattern  # noqa: N811
from re import Pattern
from re import compile as compile_re
from typing import TYPE_CHECKING, Final, Iterable, Optional, cast

from pdm_pfsc.logging import logger, traced_function

from .core.version import Pep440VersionFormatter, Version
from .vcs import HunkSource

if TYPE_CHECKING:
    from re import Match

    from .core.config import Config


# Justification: Minimal protocol. Maybe false positive,
# since 2 public methods available
class VersionSource(ABC, HunkSource):  # pylint: disable=R0903
    """"""

    @property
    @abstractmethod
    def current_version(self) -> "Version":
        """"""
        raise NotImplementedError()  # pylint: disable=R0801

    @current_version.setter
    @abstractmethod
    def current_version(self, version: "Version") -> None:
        raise NotImplementedError()  # pylint: disable=R0801

    @abstractmethod
    def get_source_file_change_hunks(self) -> list[str]:
        """"""
        raise NotImplementedError()


@dataclass
class _DiffData:
    original_lines: "list[str]" = field()
    new_lines: "list[str]" = field()

    @traced_function
    def merge(self, other: "_DiffData") -> "_DiffData":
        """"""
        return _DiffData(self.original_lines, other.new_lines)


class _VersionLine:
    """"""

    def __init__(self, prefix: str) -> None:
        """"""
        self.__prefix = prefix
        self.__version_group_name = "version"

    @traced_function
    def compile_pattern(self) -> Pattern[str]:
        """"""
        return compile_re(
            self.pattern,
            MultilinePattern,
        )

    @property
    @traced_function
    def version_group_name(self) -> str:
        """"""
        return self.__version_group_name

    @version_group_name.setter
    @traced_function
    def version_group_name(self, value: str) -> None:
        """"""
        self.__version_group_name = value

    @property
    @traced_function
    def pattern(self) -> str:
        """"""
        pattern = ""
        pattern += rf"^{self.__prefix}\s*="
        pattern += rf"\s*[\"'](?P<{self.__version_group_name}>.+?)[\"']"
        pattern += r"\s*(?:#.*)?$"
        return pattern


class _DynamicVersionConfig:
    """"""

    __file: "Path"
    __pattern: "Pattern[str]"
    __file_encoding: str

    def __init__(
        self,
        file_path: "Path",
        line_config: "_VersionLine",
        encoding: str = "utf-8",
    ) -> None:
        self.__file = file_path
        self.__line_config = line_config
        self.__file_encoding = encoding
        self.__pattern = line_config.compile_pattern()

    @cached_property
    @traced_function
    def dynamic_version(self) -> "Optional[str]":
        """"""
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            match = self.__pattern.search(file_ptr.read())
        if match is not None:
            return match.group(self.__line_config.version_group_name)
        return None

    @traced_function
    def replace_dynamic_version(self, new_version: str) -> "_DiffData":
        """

        Parameters
        ----------
        new_version: str :


        Returns
        -------

        """
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            version_file = file_ptr.read()
            match = self.__pattern.search(version_file)
            if match is None:
                raise ValueError("Failed to fetch version")
            match = cast("Match[str]", match)
            version_start, version_end = match.span(
                self.__line_config.version_group_name
            )
            new_version_file = (
                version_file[:version_start]
                + new_version
                + version_file[version_end:]
            )
        with self.__file.open("w", encoding=self.__file_encoding) as file_ptr:
            file_ptr.write(new_version_file)

        return _DiffData(
            version_file.splitlines(),
            new_version_file.splitlines(),
        )


class _VersionStorageBackend(VersionSource, ABC):
    """"""

    def __init__(self, config: "Config", repository_root: "Path") -> None:
        self.__config = config
        self.__repository_root = repository_root
        self.__current_version: "Version | None" = None
        self.__hunk: "Optional[_DiffData]" = None

    @property
    @abstractmethod
    def is_enabled(self) -> bool:
        """"""
        raise NotImplementedError()  # pylint: disable=R0801

    @property
    @abstractmethod
    def source_file_path(self) -> "Path":
        """"""
        raise NotImplementedError()  # pylint: disable=R0801

    @property
    @traced_function
    def _config(self) -> "Config":
        """"""
        return self.__config

    @property
    @traced_function
    def _repository_root(self) -> "Path":
        """"""
        return self.__repository_root

    @property
    @abstractmethod
    def line_identifier(self) -> "_VersionLine":
        """"""
        raise NotImplementedError()

    @property
    @traced_function
    def current_version(self) -> "Version":
        """"""
        if self.__current_version is None:
            dynamic: "_DynamicVersionConfig" = self._get_current_version_config
            version_line_value: "Optional[str]" = dynamic.dynamic_version
            if version_line_value is None:
                raise ValueError(
                    f"Failed to find version in {self.source_file_path}. "
                    f"Make sure it matches {self.line_identifier.pattern}"
                )
            self.__current_version = Version.from_string(version_line_value)

        return self.__current_version

    @current_version.setter
    @traced_function
    def current_version(self, new_version: "Version") -> None:
        """"""
        self.__current_version = new_version
        version_str = Pep440VersionFormatter().format(new_version)
        hunk = self._get_current_version_config.replace_dynamic_version(
            version_str
        )
        if self.__hunk is None:
            self.__hunk = hunk
        else:
            self.__hunk = cast("_DiffData", self.__hunk).merge(hunk)

    @cached_property
    @traced_function
    def _get_current_version_config(self) -> "_DynamicVersionConfig":
        """"""
        config: "_DynamicVersionConfig" = _DynamicVersionConfig(
            self.source_file_path, self.line_identifier
        )
        return config

    @traced_function
    def get_source_file_change_hunks(self) -> list[str]:
        """"""
        if self.__hunk is None:
            return []

        dynamic_version_provider: "Path" = self.source_file_path
        relative_file_path: "Path" = dynamic_version_provider.relative_to(
            str(self._repository_root)
        )
        hunk: "_DiffData" = cast("_DiffData", self.__hunk)

        return list(
            unified_diff(
                hunk.original_lines,
                hunk.new_lines,
                fromfile=str(Path("a") / relative_file_path),
                tofile=str(Path("b") / relative_file_path),
                lineterm="",
            )
        )


class _Pep621StorageBackend(_VersionStorageBackend):
    """"""

    Pattern: "Final[_VersionLine]" = _VersionLine("version")

    @property
    @traced_function
    def source_file_path(self) -> "Path":
        """"""
        return self._config.pyproject_file

    @property
    @traced_function
    def is_enabled(self) -> "bool":
        """"""
        return self._config.meta_data.version is not None

    @property
    @traced_function
    def line_identifier(self) -> "_VersionLine":
        """"""
        return self.Pattern


class _DynamicStorageBackend(_VersionStorageBackend, ABC):
    """"""

    Pattern: "Final[_VersionLine]" = _VersionLine("__version__")

    @property
    @traced_function
    def is_enabled(self) -> "bool":
        """"""
        return self._config.meta_data.is_dynamic_version


class _PdmPep517StorageBackend(_DynamicStorageBackend):
    """"""

    @property
    @traced_function
    def source_file_path(self) -> "Path":
        """"""
        build_system = self._config.meta_data.build_system
        file_path = build_system.version_source_file
        if file_path is None:
            raise ValueError(
                "Invalid configuration for deprecated PDM PEP517 configuration"
            )

        logger.warning(
            "Build backend pdm-pep517 is deprecated. Consider upgrading."
        )

        return self._repository_root / file_path

    @property
    @traced_function
    def is_enabled(self) -> "bool":
        """"""
        build_system = self._config.meta_data.build_system
        return (
            super().is_enabled
            and build_system.uses_deprecated_build_backed_pdm_pep517
        )

    @property
    @traced_function
    def line_identifier(self) -> "_VersionLine":
        """"""
        return self.Pattern


class _PdmBackendStorageBackend(_DynamicStorageBackend):
    """"""

    @property
    @traced_function
    def source_file_path(self) -> "Path":
        """"""
        build_system = self._config.meta_data.build_system
        file_path = build_system.version_source_file
        if file_path is None:
            raise ValueError(
                "Invalid configuration for PDM Backend configuration"
            )

        return self._repository_root / file_path

    @property
    @traced_function
    def is_enabled(self) -> "bool":
        """"""
        build_system = self._config.meta_data.build_system
        return (
            super().is_enabled
            and build_system.uses_pdm_backend
            and not build_system.pdm_backend.use_scm
            and build_system.pdm_backend.use_file
        )

    @property
    @traced_function
    def line_identifier(self) -> "_VersionLine":
        """"""
        return self.Pattern


@traced_function
def get_all_backends(
    config: "Config", repository_root: "Path"
) -> "Iterable[_VersionStorageBackend]":
    """"""
    yield _Pep621StorageBackend(config, repository_root)
    yield _PdmPep517StorageBackend(config, repository_root)
    yield _PdmBackendStorageBackend(config, repository_root)


@traced_function
def get_backend(
    config: "Config", repository_root: "Path"
) -> "Optional[VersionSource]":
    """"""
    for backend in get_all_backends(config, repository_root):
        if backend.is_enabled:
            return backend

    return None



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, ClassVar, Protocol

from pdm_pfsc.config import MissingValue
from pdm_pfsc.hook import HookBase, HookExecutorBase
from pdm_pfsc.logging import logger, traced_function

from ..core.config import (
    ALLOW_DIRTY_DEFAULT,
    AUTO_TAG_DEFAULT,
    COMMIT_MESSAGE_TEMPLATE_DEFAULT,
    PERFORM_COMMIT_DEFAULT,
    TAG_ADD_PREFIX_DEFAULT,
)
from ..core.version import Pep440VersionFormatter, Version

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace
    from typing import Any

    from ..vcs import HunkSource, VcsProvider


def _str_as_bool(match: "Any") -> bool:
    match_str = str(match)
    false_strings = ("false",)

    if match_str.lower() in false_strings:
        return False

    return bool(match)


class _Executable(Protocol):  # pylint: disable=R0903
    # Justification: Just a protocol
    """"""

    def run(self, dry_run: bool = False) -> "Version":
        """

        Parameters
        ----------
            dry_run :


        Returns
        -------

        """
        raise NotImplementedError()


@dataclass(frozen=True)
class PreHookContext:
    """"""

    vcs_provider: "VcsProvider" = field()
    version: "Version" = field()

    @cached_property
    def formatted_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.version)


@dataclass(frozen=True)
class PostHookContext:
    """"""

    vcs_provider: "VcsProvider" = field()
    hunk_source: "HunkSource" = field()
    version: "Version" = field()
    previous_version: "Version" = field()
    version_changed: bool = field()

    @cached_property
    def formatted_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.version)

    @cached_property
    def formatted_previous_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.previous_version)


class HookExecutor(HookExecutorBase[tuple[_Executable, Version]]):
    """"""

    def __init__(
        self, hunk_source: "HunkSource", vcs_provider: "VcsProvider"
    ) -> None:
        """

        Parameters:
        -----------
            hunk_source: HunkSource :
            vcs_provider: VcsProvider :

        Returns:
        --------

        """
        self.__hunk_source = hunk_source
        self.__vcs_provider = vcs_provider
        super().__init__()

    @traced_function
    def run(
        self,
        context: "tuple[_Executable, Version]",
        args: "Namespace",
        dry_run: bool = False,
    ) -> None:
        """

        Parameters:
        -----------
            context: tuple[_Executable, Version] :
            args: Namespace :
            dry_run: bool :

        Returns:
        --------

        """
        (executor, version) = context

        pre_call_ctx: "PreHookContext" = PreHookContext(
            self.__vcs_provider, version
        )

        if not dry_run:
            for hook in self._hooks:
                hook.pre_action_hook(pre_call_ctx, args)

        old_version = version
        version = executor.run(dry_run)

        post_call_ctx: "PostHookContext" = PostHookContext(
            self.__vcs_provider,
            self.__hunk_source,
            version,
            old_version,
            old_version != version,
        )

        if not dry_run:
            for hook in self._hooks:
                hook.post_action_hook(post_call_ctx, args)


class CommitChanges(HookBase):
    """"""

    default_commit_message: "ClassVar[str]" = COMMIT_MESSAGE_TEMPLATE_DEFAULT
    perform_commit: "ClassVar[bool]" = PERFORM_COMMIT_DEFAULT

    @traced_function
    def post_action_hook(
        self, context: "PostHookContext", args: "Namespace"
    ) -> None:
        """

        Parameters:
        -----------
            context: PostHookContext :
            args: Namespace :

        Returns:
        --------

        """
        kwargs = vars(args)
        if not context.version_changed:
            logger.debug("Cannot commit changes. Nothing to commit")
            return

        if _str_as_bool(kwargs.pop("commit", self.perform_commit)):
            message: str = str(
                kwargs.pop("commit_message", self.default_commit_message)
            )

            f_args: dict[str, str] = {
                "to": context.formatted_version,
                "from": context.formatted_previous_version,
            }

            message = message.format(**f_args)

            context.vcs_provider.check_in_deltas(message, context.hunk_source)

    @traced_function
    def pre_action_hook(
        self, context: "PreHookContext", args: "Namespace"
    ) -> None:
        """

        Parameters:
        -----------
            context: PreHookContext :
            args: Namespace :

        Returns:
        --------

        """
        return  # NOSONAR

    @classmethod
    @traced_function
    def configure(cls, parser: "ArgumentParser") -> None:
        """

        Parameters:
        -----------

        parser: ArgumentParser :

        Returns:
        --------

        """
        parser.add_argument(
            "--commit",
            "-c",
            action="store_true",
            default=MissingValue(CommitChanges.perform_commit),
            help="Commit changes to repository. Uses configuration value "
            "'perform_commit' to store default action.",
        )

        parser.add_argument(
            "--message",
            "-m",
            dest="commit_message",
            action="store",
            default=MissingValue(CommitChanges.default_commit_message),
            help="The commit message template. May contain "
            "{from} and {to} as format identifier. Uses configuration value "
            "'commit_msg_tmpl' as configured default value.",
        )


class TagChanges(HookBase):
    """"""

    do_tag: "ClassVar[bool]" = AUTO_TAG_DEFAULT
    allow_dirty: "ClassVar[bool]" = ALLOW_DIRTY_DEFAULT
    prepend_to_tag: "ClassVar[bool]" = TAG_ADD_PREFIX_DEFAULT

    @traced_function
    def post_action_hook(
        self, context: "PostHookContext", args: "Namespace"
    ) -> None:
        """

        Parameters:
        -----------

            context: PostHookContext :

            args: Namespace :

        Returns:
        --------

        """
        kwargs = vars(args)
        if not context.version_changed:
            return

        tag_repo = _str_as_bool(kwargs.pop("tag", self.do_tag))
        must_be_clean = not _str_as_bool(kwargs.pop("dirty", self.allow_dirty))
        is_dirty = not context.vcs_provider.is_clean

        if tag_repo:
            if must_be_clean and is_dirty:
                raise RuntimeError(
                    "This should only take place, if "
                    "the git repository is clean"
                )
            if is_dirty:
                logger.warning(
                    "The repository is not clean. Performing tag anyway."
                )

            context.vcs_provider.create_tag_from_version(
                context.version,
                _str_as_bool(
                    kwargs.pop("prepend_letter_v", self.prepend_to_tag)
                ),
            )

    @traced_function
    def pre_action_hook(
        self, context: "PreHookContext", args: "Namespace"
    ) -> None:
        """

        Parameters:
        -----------

            context: PreHookContext :

            args: Namespace :

        Returns:
        --------

        """
        kwargs = vars(args)
        if (
            _str_as_bool(kwargs.get("tag", self.do_tag))
            and not _str_as_bool(kwargs.get("dirty", self.allow_dirty))
            and not context.vcs_provider.is_clean
        ):
            raise RuntimeError("Repository root is not clean")

    @classmethod
    @traced_function
    def configure(cls, parser: "ArgumentParser") -> None:
        """
        Parameters:
        -----------
            parser: ArgumentParser :

        Returns:
        --------

        """
        parser.add_argument(
            "--tag",
            "-t",
            action="store_true",
            default=MissingValue(TagChanges.do_tag),
            help="Create a tag after modifying the current version. Uses "
            "configuration value 'auto_tag' to store its default value.",
        )

        parser.add_argument(
            "--dirty",
            "-d",
            action="store_true",
            default=MissingValue(TagChanges.allow_dirty),
            help="Create a tag, even if the repository is dirty. Uses "
            "configuration value 'allow_dirty' to store its "
            "default value.",
        )

        parser.add_argument(
            "--no-prepend-v",
            dest="prepend_letter_v",
            action="store_false",
            default=MissingValue(TagChanges.prepend_to_tag),
            help="Do not prepend letter v for the tag. Uses "
            "configuration value 'tag_add_prefix' to store its "
            "default value.",
        )
